/* 
	FILE: Surface.cpp
	PURPOSE: NURBS and Bezier surfaces main class (implementation)
	Author: Carlos Jimnez de Parga (c)
	License: GPL v3.0
*/

#include "surface.h"

// ----------------------------------------------------------------------------
// Control Point class implementation
// ----------------------------------------------------------------------------


CtrlPoint::CtrlPoint(GLuint name,float ptx, float pty, float ptz, float angleX, float angleY, float width, float height):VertexBox(ptx, pty, ptz, width, height)
{
	m_Name = name;

	m_xAngle = angleY;
	m_yAngle = angleX;

	m_CorX = m_CorY = false;

}

CtrlPoint::CtrlPoint(GLuint name)
{
	m_Name = name;
}

// Draws selection box

void CtrlPoint::DrawBox()
{

	glPushMatrix();

	glEnable(GL_LINE_STIPPLE);
	glLineStipple(5,0xaaaa);
	glTranslatef(m_tx,m_ty,m_tz);
	glColor3f(1.0f,1.0f,0.0f);
	glBegin(GL_QUADS);
	glVertex3f(-1.0f, -1.0f, 0.0f);
	glVertex3f(m_w + 1.0f, -1.0f,0.0f);
	glVertex3f(m_w + 1.0f, m_h + 1.0f, 0.0f);
	glVertex3f(-1.0f, m_h + 1.0f, 0.0f);
	glEnd();
	glDisable(GL_LINE_STIPPLE); 

	glPopMatrix();

}

// Draws control point

void CtrlPoint::Draw()
{
	glPushMatrix();

	glTranslatef(m_tx,m_ty,m_tz);
	glLineWidth(2);
	glLoadName(m_Name);
	glBegin(GL_QUADS);
	glColor3f(0.1f,0.1f,1.0f);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glVertex3f(m_w, 0.0f, 0.0f);
	glVertex3f(m_w, m_h, 0.0f);
	glVertex3f(0.0f, m_h, 0.0f);
	glEnd();
	glLineWidth(1);
	glPopMatrix();
}

// Accessors

Tool::Vector3D CtrlPoint::GetVectorT()
{
	return Tool::Vector3D(m_tx,m_ty,m_tz);
}

Tool::Vector3D CtrlPoint::GetVectorM()
{
	return Tool::Vector3D(m_x,m_y,m_z);
}


void CtrlPoint::SetVector(Tool::Vector3D Vector)
{
	m_tx = Vector.m_x;
	m_ty = Vector.m_y;
	m_tz = Vector.m_z;
	SetPassV(true); // If surface moved then m_PassV and m_PassR = true
	SetPassR(true); // This will generate a new m_x, m_y and m_z according the new m_tx, m_ty and m_tz after translation
}


// ----------------------------------------------------------------------------
// Surface class implementation
// ----------------------------------------------------------------------------


Surface::Surface(int Limit):VertexBox()
{
	m_Limit = Limit;
	m_LimitCount = 0;
	m_MaxX = -9999999.0f;
	m_MaxY = -9999999.0f;
	m_MinX =  9999999.0f;
	m_MinY =  9999999.0f;
	m_OldTX = 0.0f;
	m_OldTY = 0.0f;
	m_MkAvg = true;
	m_Pass = true;
	m_ControlPointList = NULL;
	m_RotState = ID_ROT_NOTHING;
}

// Performs copy

void Surface::Copy(Surface* s)
{
	s->DoCopy(m_LimitCount, m_OldTX, m_OldTY, m_tx, m_ty, m_MaxX, m_MaxY, m_MinX, m_MinY, m_ControlPointList);
	VertexBox::Copy(s);
}

void Surface::DoCopy(int LimitCount, float OldTX, float OldTY, float tx, float ty, float MaxX, float MaxY, float MinX, float MinY, ControlPoint* cp)
{
	m_LimitCount = LimitCount;
	m_OldTX = (m_OldTX == 0.0f ? tx : OldTX);
	m_OldTY = (m_OldTY == 0.0f ? ty : OldTY);
	m_MaxX = MaxX;
	m_MaxY = MaxY;
	m_MinX = MinX;
	m_MinY = MinY;
	m_Pass = true;

	unsigned int CountCtrlCP = m_Name + 1;

	for(ControlPoint::iterator it = cp->begin();it != cp->end();it++)
	{
		CtrlPoint *cpC = *it;
		CtrlPoint *ncp = new CtrlPoint(CountCtrlCP++);
		cpC->VertexBox::Copy(ncp);
		m_ControlPointList->Append(ncp);
	}

}

// Increments number of control points

void Surface::Increment()
{
	m_LimitCount++;
}

// Decrements number of control points

void Surface::Decrement()
{
	m_LimitCount--;
}

void Surface::SetMkAverage(bool Avg)
{
	m_MkAvg = Avg;
}

void Surface::SetRotState(StateRot State)
{
	m_RotState = State;
}

Surface::StateRot Surface::GetRotState()
{
	return m_RotState;
}

// Accessors

void Surface::SetCtrlPoints(ControlPoint* CtrlPts)
{
	m_ControlPointList = CtrlPts;
}

ControlPoint* Surface::GetCtrlPoints()
{
	return m_ControlPointList;
}


// Go through the control list object and update the control list array

void Surface::ProcessCtrlPoints()
{
	int i = 0,j = 0,k = 0;

	m_MaxX = -9999999.0f;
	m_MaxY = -9999999.0f;
	m_MinX =  9999999.0f;
	m_MinY =  9999999.0f;


	for(ControlPoint::iterator it = m_ControlPointList->begin();it != m_ControlPointList->end();it++)
	{
		CtrlPoint *cp = *it;
		m_CtrlPoint[i][j][k] = cp->GetVectorT().m_x;
		m_CtrlPoint[i][j][k+1] = cp->GetVectorT().m_y;
		m_CtrlPoint[i][j][k+2] = cp->GetVectorT().m_z;

		if (cp->GetVectorT().m_x > m_MaxX) 
			m_MaxX = cp->GetVectorT().m_x;
		if (cp->GetVectorT().m_y > m_MaxY) 	
			m_MaxY = cp->GetVectorT().m_y;
		if (cp->GetVectorT().m_x < m_MinX) 
			m_MinX = cp->GetVectorT().m_x;
		if (cp->GetVectorT().m_y < m_MinY) 
			m_MinY = cp->GetVectorT().m_y;

		j++;
		if (j == 8) {
			j = 0;
			i++;
		}
	}

	m_w  = m_MaxX;
	m_h  = m_MaxY;
	m_tx = m_MinX;
	m_ty = m_MinY;
	m_tz = 0.0f;

	m_x = m_MinX;
	m_y = m_MaxY;
}

// Calculates control points average. Gravitational center of the surface.

void Surface::MakeAllAverage()
{
	if (m_MkAvg)
	{
	
		Tool::Vector3D Average;
	
		for(ControlPoint::iterator it = m_ControlPointList->begin();it != m_ControlPointList->end();it++)
		{
			CtrlPoint *cp = *it;
			
			Average.m_x+= cp->GetVectorT().m_x;
			Average.m_y+= cp->GetVectorT().m_y;
			Average.m_z+= cp->GetVectorT().m_z;
		}

		Average.m_x /= m_ControlPointList->GetCount();
		Average.m_y /= m_ControlPointList->GetCount();
		Average.m_z /= m_ControlPointList->GetCount();

		for(ControlPoint::iterator it = m_ControlPointList->begin();it != m_ControlPointList->end();it++)
		{
			CtrlPoint *cp = *it;
			cp->SetAverage(Average); // Informs each control point
		}
	}
}


// Checks the control list's limit

bool Surface::CheckLimit()
{
	return (m_LimitCount == m_Limit);
}

// Reset the first iteration in the move algorithm

void Surface::ResetPass()
{
	m_Pass = false;
}

void Surface::Spin(float xSpin, float ySpin)
{
	m_xAngle+= xSpin; 
	m_yAngle+= ySpin;

	m_xAngle %= 360;
	m_yAngle %= 360;
}

// Performs a complete move of all surface

void Surface::Move(int x, int y, float SceneX, float SceneY, int ScreenWidth, int ScreenHeight)
{

	float auxX = Tool::Math::GetOGLPos(x,y, ScreenWidth, ScreenHeight).m_x + SceneX;
	float auxY = Tool::Math::GetOGLPos(x,y, ScreenWidth, ScreenHeight).m_y + SceneY;
	
	if (!m_Pass) // Check for first iteration
	{
		m_OldTX = auxX;
		m_OldTY = auxY;
		m_Pass = true;
	}

	Tool::Vector3D Vector;
	
		
	float incX=0.0f;
	float incY=0.0f;
	
	// Calculates DeltaX(incX) and DeltaY(incY) for all ctrl points displacement

	for(ControlPoint::iterator it = m_ControlPointList->begin();it != m_ControlPointList->end();it++){
		CtrlPoint *cp = *it;
		Vector.m_x = cp->GetVectorT().m_x;
		Vector.m_y = cp->GetVectorT().m_y;

		incX = Vector.m_x + (auxX - m_OldTX); // Delta increment
		incY = Vector.m_y + (auxY - m_OldTY);

		cp->SetVector(Tool::Vector3D(incX,incY,cp->GetVectorT().m_z));
	}

	// Updates last positions

	m_OldTX = auxX;
	m_OldTY = auxY;

	m_MaxX = -9999999.0f;
	m_MaxY = -9999999.0f;
	m_MinX =  9999999.0f;
	m_MinY =  9999999.0f;

}


// Destroy the surface

void Surface::Destroy()
{
	if (m_ControlPointList)
		for(ControlPoint::iterator it = m_ControlPointList->begin();it != m_ControlPointList->end();it++)
		{
			CtrlPoint *cp = *it;
			delete cp;
		}

		delete m_ControlPointList;
}

// Save part to XML

void Surface::Save(wxString& XMLContent)
{

	if (m_ControlPointList)
	{

		XMLContent+=_("<issisurface>");
		XMLContent+= _("<m_LimitCount>") + _(Tool::Math::ConvertNumber(m_LimitCount)) +  _("</m_LimitCount>");
		XMLContent+= _("<m_OldTX>") + _(Tool::Math::ConvertNumber(m_OldTX)) +  _("</m_OldTX>");
		XMLContent+= _("<m_OldTY>") + _(Tool::Math::ConvertNumber(m_OldTY)) +  _("</m_OldTY>");
		XMLContent+= _("<m_MaxX>") + _(Tool::Math::ConvertNumber(m_MaxX)) +  _("</m_MaxX>");
		XMLContent+= _("<m_MaxY>") + _(Tool::Math::ConvertNumber(m_MaxY)) +  _("</m_MaxY>");
		XMLContent+= _("<m_MinX>") + _(Tool::Math::ConvertNumber(m_MinX)) +  _("</m_MinX>");
		XMLContent+= _("<m_MinY>") + _(Tool::Math::ConvertNumber(m_MinY)) +  _("</m_MinY>");
		XMLContent+= _("<m_Pass>") + _(Tool::Math::ConvertNumber(m_Pass)) +  _("</m_Pass>");
		const int Red = m_Color.Red();
		XMLContent+= _("<m_Red>") + _(Tool::Math::ConvertNumber(Red)) + _("</m_Red>");
	
		const int Green = m_Color.Green();
		XMLContent+= _("<m_Green>") + _(Tool::Math::ConvertNumber(Green)) + _("</m_Green>");
	
		const int Blue = m_Color.Blue();
		XMLContent+= _("<m_Blue>") + _(Tool::Math::ConvertNumber(Blue)) + _("</m_Blue>");

		XMLContent+= _("<m_Alpha>") + _(Tool::Math::ConvertNumber(m_Alpha)) + _("</m_Alpha>");
			
		XMLContent+=_("</issisurface>");

		for(ControlPoint::iterator it = m_ControlPointList->begin();it != m_ControlPointList->end();it++)
		{
			XMLContent+=_("<issicontrolpoint>");
			(*it)->Save(XMLContent);
			XMLContent+=_("</issicontrolpoint>");		
		}
	}
}

// Load part from XML

unsigned int Surface::CtrlCPLoad(xml_node<>* Node)
{

	if (!strcmp(Node->name(),"issisurface"))
	{
		xml_node<>* InNode = Node->first_node();

		m_LimitCount = atoi(InNode->value());
		InNode = InNode->next_sibling();

		m_OldTX = atof(InNode->value());
		InNode = InNode->next_sibling();

		m_OldTY = atof(InNode->value());
		InNode = InNode->next_sibling();

		m_MaxX = atof(InNode->value());
		InNode = InNode->next_sibling();

		m_MaxY = atof(InNode->value());
		InNode = InNode->next_sibling();

		m_MinX = atof(InNode->value());
		InNode = InNode->next_sibling();

		m_MinY = atof(InNode->value());
		InNode = InNode->next_sibling();

		m_Pass = atoi(InNode->value());
		InNode = InNode->next_sibling();
		
		int Red;
		Red = atof(InNode->value());
		InNode = InNode->next_sibling();

		int Green;
		Green = atof(InNode->value());
		InNode = InNode->next_sibling();

		int Blue;
		Blue = atof(InNode->value());
		InNode = InNode->next_sibling();
	
		m_Color = wxColor(Red, Green, Blue);

		int Alpha;
		Alpha = atof(InNode->value());
	
		m_Alpha = Alpha;
	}

	Node = Node->next_sibling();	

	unsigned int CountCtrlCP = m_Name + 1;

	for(xml_node<>* node = Node; node; node = node->next_sibling())
	{
		if (!strcmp(node->first_node()->name(),"issivertex"))
		{
			CtrlPoint *ncp = new CtrlPoint(CountCtrlCP++);
			m_ControlPointList->Append(ncp);
			ncp->Load(node->first_node());		
			ncp->Object::Load(node->first_node()->next_sibling()->first_node());
		}

	}
	return CountCtrlCP;
}

// Destructor

Surface::~Surface()
{
	Destroy();
}

// ----------------------------------------------------------------------------
// NURBS class implementation
// ----------------------------------------------------------------------------


NURBS::NURBS(int Limit, GLuint name):Surface(Limit)
{
	m_KnotsU = NULL;
	
	GLfloat Knots[16] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0};
	memcpy(m_KnotsV,Knots,sizeof(GLfloat) * 16);

	
	m_UpToRender = 0;
	m_OldCount = -1;
	m_Nurb = gluNewNurbsRenderer();
	gluNurbsProperty(m_Nurb, GLU_DISPLAY_MODE, GLU_OUTLINE_POLYGON);
	m_Name = name;
}

// Draws the selction box around the Nurbs surface

void NURBS::DrawBox()
{

	glPushMatrix();

	glEnable(GL_LINE_STIPPLE);
	glLineStipple(5,0xaaaa);

	glColor3f(1.0f,1.0f,0.0f);
	glTranslatef(m_tx,m_ty,m_tz);
	glBegin(GL_QUADS);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glVertex3f(abs(m_w - m_tx), 0.0f, 0.0f);
		glVertex3f(abs(m_w - m_tx), abs(m_h - m_ty), 0.0f);
		glVertex3f(0.0f, abs(m_h - m_ty), 0.0f);
	glEnd();

	glDisable(GL_LINE_STIPPLE); 

	glPopMatrix();

}

// Recalculate knots when deleted

void NURBS::RecalculateDeletedCP()
{
	if (m_ControlPointList->GetCount() + 1 == m_UpToRender)
		m_UpToRender-=8;

	GenerateKnots();
	m_OldCount = m_ControlPointList->GetCount();
}


// Draws the surface

void NURBS::Draw()
{

	if (m_ControlPointList->GetCount() % 8 == 0)
	{
		m_UpToRender = m_ControlPointList->GetCount();

		if (m_OldCount != m_ControlPointList->GetCount()) // PERFORMANCE ISSUE: For not generating calculus each loop
		{
			GenerateKnots();
			m_OldCount = m_ControlPointList->GetCount();
		
		}
	}

	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);
	gluBeginSurface(m_Nurb);
	
	gluNurbsSurface(m_Nurb, m_UpToRender / 4, m_KnotsU, 16,  m_KnotsV, 24, 3, &m_CtrlPoint[0][0][0], 
		m_UpToRender / 8 , 8 , GL_MAP2_VERTEX_3);

	gluEndSurface(m_Nurb);

}

// Draws the selection handler

void NURBS::DrawSelection()
{

	glPushMatrix();
	glColor3f(1.0f ,0.5f,0.5f);
	glLoadName(m_Name);	
	glTranslatef(m_tx,m_ty,m_tz);
	glBegin(GL_QUADS);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glVertex3f(2.0f, 0.0f, 0.0f);
		glVertex3f(2.0f, 2.0f, 0.0f);
		glVertex3f(0.0f, 2.0f, 0.0f);
	glEnd();

	glPopMatrix();
}

// Destory nurbs

void NURBS::Destroy()
{

	if (m_KnotsU)
		delete [] m_KnotsU;

	if (m_Nurb)
		gluDeleteNurbsRenderer(m_Nurb);

}

// Ready for display Nurbs (when count >=8 control points)

bool NURBS::GetReady()
{
	if (m_ControlPointList->GetCount() >= 8) return true;
	else return false;

}

// Calculate knots

void NURBS::GenerateKnots()
{

	const int nU = m_UpToRender / 4;

	if (m_KnotsU) delete [] m_KnotsU;

	m_KnotsU = new GLfloat[nU];


	for(int i = 0; i < nU / 2; i++)
		m_KnotsU[i] = 0.0f;

	for(int i = nU / 2; i < nU; i++)
		m_KnotsU[i] = 1.0f;

	

}

// Performs copy

void NURBS::Copy(NURBS* Nurbs)
{
	Nurbs->DoCopy(m_UpToRender);
	Surface::Copy(Nurbs);	
	Nurbs->GenerateKnots();
}

void NURBS::DoCopy(int UpToRender)
{
	m_UpToRender = UpToRender;
}

// Save part to XML

void NURBS::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"Nurbs\">");
	XMLContent+=_("<issinurbs>");
	XMLContent+= _("<m_UpToRender>") + _(Tool::Math::ConvertNumber(m_UpToRender)) +  _("</m_UpToRender>");
	XMLContent+= _("<m_OldCount>") + _(Tool::Math::ConvertNumber(m_OldCount)) +  _("</m_OldCount>");
	XMLContent+=_("</issinurbs>");
	Surface::Save(XMLContent);
	XMLContent+=_("</issiform>");
}

// Load part from XML

void NURBS::Load(xml_node<>* Node)
{
	xml_node<>* InNode = Node->first_node();

	m_UpToRender = atoi(InNode->value());
	InNode = InNode->next_sibling();

	m_OldCount = atoi(InNode->value());

	Surface::Load(Node->next_sibling());

	GenerateKnots();
}

NURBS::~NURBS()
{
	Destroy();
}


// ----------------------------------------------------------------------------
// NURBS class implementation
// ----------------------------------------------------------------------------


BEZIER::BEZIER(int Limit, GLuint name):Surface(Limit)
{
	m_Name = name;
}

// Draws the selction box around the Nurbs surface

void BEZIER::DrawBox()
{

	glPushMatrix();

	glEnable(GL_LINE_STIPPLE);
	glLineStipple(5,0xaaaa);

	glColor3f(1.0f,1.0f,0.0f);
	glTranslatef(m_tx,m_ty,m_tz);
	glBegin(GL_QUADS);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glVertex3f(abs(m_w - m_tx), 0.0f, 0.0f);
		glVertex3f(abs(m_w - m_tx), abs(m_h - m_ty), 0.0f);
		glVertex3f(0.0f, abs(m_h - m_ty), 0.0f);
	glEnd();

	glDisable(GL_LINE_STIPPLE); 

	glPopMatrix();

}

// Draws the surface

void BEZIER::Draw()
{

	glColor4f(m_Color.Red() / 255.0f ,m_Color.Green() / 255.0f, m_Color.Blue() / 255.0f, m_Alpha/ 100.0f);

	glMap2f(GL_MAP2_VERTEX_3, 0.0f, 10.0f, 3, 8, 0.0f,
		10.0f, 24, m_ControlPointList->GetCount() / 8, &m_CtrlPoint[0][0][0]);

	glEnable(GL_MAP2_VERTEX_3);

	glMapGrid2f(20, 0.0, 10.0, 20, 0.0, 10.0);

	glEvalMesh2(GL_LINE, 0, 20, 0, 20);

	glDisable(GL_MAP2_VERTEX_3);


}

// Draws the selection handler

void BEZIER::DrawSelection()
{

	glPushMatrix();
	glColor3f(1.0f ,0.5f,0.5f);
	glLoadName(m_Name);	
	glTranslatef(m_tx,m_ty,m_tz);
	glBegin(GL_QUADS);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glVertex3f(2.0f, 0.0f, 0.0f);
		glVertex3f(2.0f, 2.0f, 0.0f);
		glVertex3f(0.0f, 2.0f, 0.0f);
	glEnd();

	glPopMatrix();
}

// Rady for display Bezier when control points >= 4

bool BEZIER::GetReady()
{
	if (m_ControlPointList->GetCount() >= 8) return true;
	else return false;

}

// Save part to XML

void BEZIER::Save(wxString& XMLContent)
{
	XMLContent+=_("<issiform form = \"Bezier\">");
	Surface::Save(XMLContent);
	XMLContent+=_("</issiform>");
}


